theory Demo09
  imports Solutions07
begin

subsection \<open>Well-Foundedness\<close>

(*infinite decreasing chains*)
definition "ichain r f \<longleftrightarrow> (\<forall>i. (f (Suc i), f i) \<in> r)"

(*admits induction*)
definition "inductive r \<longleftrightarrow> (\<forall>P x. (\<forall>x. (\<forall>y. (y, x) \<in> r \<longrightarrow> P y) \<longrightarrow> P x) \<longrightarrow> P x)"

(*accessible part of relation*)
inductive_set accessible :: "('a \<times> 'a) set \<Rightarrow> 'a set" for r :: "('a \<times> 'a) set"
  where
    "(\<And>y. (y, x) \<in> r \<Longrightarrow> y \<in> accessible r) \<Longrightarrow> x \<in> accessible r"

(*all non-empty sets have a minimal element*)
definition "minimal r \<longleftrightarrow> (\<forall>Q. Q \<noteq> {} \<longrightarrow> (\<exists>z\<in>Q. \<forall>y. (y, z) \<in> r \<longrightarrow> y \<notin> Q))"

(*set of nonterminating sets*)
definition "Nonterm r = {A. A \<noteq> {} \<and> (\<forall>x\<in>A. \<exists>y\<in>A. (y, x) \<in> r)}"

lemma not_inductive_imp_ichain:
  assumes "\<not> inductive r"
  shows "\<exists>f. ichain r f"
proof -
  from assms obtain P and x where "\<And>x. (\<And>y. (y, x) \<in> r \<Longrightarrow> P y) \<Longrightarrow> P x"
    and "\<not> P x" by (auto simp: inductive_def)
  then have "(\<forall>z. \<not> P z \<longrightarrow> (\<exists>y. (y, z) \<in> r \<and> \<not> P y))" by blast
  then obtain g where *: "\<forall>z. \<not> P z \<longrightarrow> \<not> P (g z) \<and> (g z, z) \<in> r" by metis
  define f where "f i = (g ^^ i) x" for i
  have "\<not> P (f i)" and "(f (Suc i), f i) \<in> r" for i
    using \<open>\<not> P x\<close> and * by (induction i) (auto simp: f_def)
  then show ?thesis by (auto simp: ichain_def)
qed

lemma no_ichain_imp_inductive:
  assumes "\<nexists>f. ichain r f"
  shows "inductive r"
  using assms and not_inductive_imp_ichain by blast

lemma inductive_imp_accessible:
  assumes "inductive r"
  shows "x \<in> accessible r"
  sorry

lemma All_accessible_imp_minimal:
  assumes "\<forall>x. x \<in> accessible r"
  shows "minimal r"
proof (unfold minimal_def, intro allI impI)
  fix Q :: "'a set" assume "Q \<noteq> {}"
  then obtain x where "x \<in> accessible r" and "x \<in> Q" using assms by auto
  then show "\<exists>z\<in>Q. \<forall>y. (y, z) \<in> r \<longrightarrow> y \<notin> Q" by (induction) auto
qed

lemma minimal_imp_Nonterm_empty:
  assumes "minimal r"
  shows "Nonterm r = {}"
proof (rule ccontr)
  assume "Nonterm r \<noteq> {}"
  then obtain N where "N \<in> Nonterm r" and "N \<noteq> {}" by (auto simp: Nonterm_def)
  with assms have "\<exists>z\<in>N. \<forall>y. (y, z) \<in> r \<longrightarrow> y \<notin> N" by (auto simp: minimal_def)
  with \<open>N \<in> Nonterm r\<close> show False by (auto simp: Nonterm_def)
qed

lemma ichain_imp_Nonterm:
  assumes "ichain r f"
  shows "f ` UNIV \<in> Nonterm r"
  sorry

lemma Nonterm_empty_imp_no_ichain:
  assumes "Nonterm r = {}"
  shows "\<nexists>f. ichain r f"
  using assms and ichain_imp_Nonterm [of r] by blast

lemma
  shows CI: "(\<nexists>f. ichain r f) \<longleftrightarrow> inductive r"
    and IA: "inductive r \<longleftrightarrow> (\<forall>x. x \<in> accessible r)"
    and AM: "(\<forall>x. x \<in> accessible r) \<longleftrightarrow> minimal r"
    and MN: "minimal r \<longleftrightarrow> Nonterm r = {}"
  using no_ichain_imp_inductive [of r]
    and inductive_imp_accessible [of r]
    and All_accessible_imp_minimal [of r]
    and minimal_imp_Nonterm_empty [of r]
    and Nonterm_empty_imp_no_ichain [of r]
  by blast+

lemma wf_iff_inductive : "wf r \<longleftrightarrow> inductive r"
  by (unfold wf_def inductive_def) blast


subsection \<open>Manual Termination Proofs\<close>

(*
fun group :: "'a list \<Rightarrow> 'a list list"
  where
    "group [] = []"
  | "group (x # xs) = (let (ys, zs) = split_same x xs in (x # ys) # group zs)"

lemma "unstutter xs = map hd (group xs)"
  sorry
*)

(*
fun evenodd :: "nat \<Rightarrow> bool \<Rightarrow> bool"
  where
    "evenodd 0 True \<longleftrightarrow> True"
  | "evenodd x False \<longleftrightarrow> \<not> (evenodd x True)"
  | "evenodd (Suc x) True \<longleftrightarrow> evenodd x False"
*)

datatype tree = Tree nat "tree list"

fun map :: "('a \<Rightarrow> 'b) \<Rightarrow> 'a list \<Rightarrow> 'b list"
  where
    "map f [] = []"
  | "map f (x#xs) = f x # map f xs"

(*
fun mirror :: "tree \<Rightarrow> tree"
  where
    "mirror (Tree n ts) = Tree n (rev (map mirror ts))"
*)

end
